/*
    Filter->de- for VirtualDub (0.09).
		Based on Red/Green/Blue (RGB) Adjustment Filter for VirtualDub.(1.0) by Donald A. Graft
    Copyright (C) 1999-2000 Donald A. Graft
	Copyright (C) 2004 Eugene Khoroshavin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	Eugene Khoroshavin
	fdump@mail.ru
*/

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <crtdbg.h>
#include <math.h>
#include <stdlib.h>

#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#include "resource.h"
#include "filter.h"

///////////////////////////////////////////////////////////////////////////

int RunProc(const FilterActivation *fa, const FilterFunctions *ff);
int StartProc(FilterActivation *fa, const FilterFunctions *ff);
int EndProc(FilterActivation *fa, const FilterFunctions *ff);
long ParamProc(FilterActivation *fa, const FilterFunctions *ff);
int InitProc(FilterActivation *fa, const FilterFunctions *ff);
int ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str);
void ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);
int normpos( int a, int b, int c, int pos, long width);
void normrgb(long *r, long *g, long *b);
void normc(long *c);
int normv(int v, int norm);
///////////////////////////////////////////////////////////////////////////
#define u22 (16384)
#define TTH (10000)

const char * f11txt = "Red channel Shift";
const char * f12txt = "Green channel Shift";
const char * f13txt = "Blue channel Shift";
const char * f1ntxt = "Chroma Shift";
const char * f1ef1txt = "Pt chain Shift";
const char * f1ef2txt = "hor. Shift";
const char * f1ef3txt = "ver. Shift";




typedef struct MyFilterData {
	IFilterPreview		*ifp;

	int					rshpt;
	int					rshxx;
	int					rshyy;

	int					gshpt;
	int					gshxx;
	int					gshyy;

	int					bshpt;
	int					bshxx;
	int					bshyy;

	int					rshxy;
	int					gshxy;
	int					bshxy;

	int					shiFlg;
	long				salign;
	long				dalign;
	int					keeppt;

	int					rshlock;
	int					gshlock;
	int					bshlock;
	int					shiunion;
	int					nmfunc;
	long				width;
	long				height;
} MyFilterData;

////////////////////////////////////////////////////////////////////////////////
void setFlg (MyFilterData *mfd);
void shiftch(MyFilterData *mfd, long *r, long *g, long *b, Pixel32 *src,  int x, int y);
///////////////////////////////////////////////////////////////////////////////

bool FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	_snprintf(buf, buflen, "Config(%d, %d, %d, %d, %d, %d)",
								  //1	2	3   4   5   6
		(mfd->rshpt + 100)	| ((mfd->rshxx + 100) << 8),	//1
		(mfd->rshyy + 100)	| ((mfd->gshpt + 100) << 8),	//2
		(mfd->gshxx + 100)	| ((mfd->gshyy + 100) << 8),	//3
		(mfd->bshpt + 100)	| ((mfd->bshxx + 100) << 8),	//4
		(mfd->bshyy + 100)	,								//5
		(mfd->keeppt)			|
		(mfd->rshlock	<< 1)	|
		(mfd->gshlock	<< 2)	|
		(mfd->bshlock	<< 3)	|
		(mfd->shiunion	<< 4)	);							//6

	return true;
}

void ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
	FilterActivation *fa = (FilterActivation *)lpVoid;
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	mfd->rshpt			= ( argv[0].asInt() & 0xff) - 100;
	mfd->rshxx			= ((argv[0].asInt() & 0xff00) >>8) -100;
	mfd->rshyy			= ( argv[1].asInt() & 0xff) - 100;
	mfd->gshpt			= ((argv[1].asInt() & 0xff00) >>8) -100;
	mfd->gshxx			= ( argv[2].asInt() & 0xff) - 100;
	mfd->gshyy			= ((argv[2].asInt() & 0xff00) >>8) -100;
	mfd->bshpt			= ( argv[3].asInt() & 0xff) - 100;
	mfd->bshxx			= ((argv[3].asInt() & 0xff00) >>8) -100;
	mfd->bshyy			= ( argv[4].asInt() & 0xff) - 100;

	int flgs = argv[5].asInt();
	mfd->keeppt			= ((flgs & 0x0001) > 0);
	mfd->rshlock		= ((flgs & 0x0002) > 0);
	mfd->gshlock		= ((flgs & 0x0004) > 0);
	mfd->bshlock		= ((flgs & 0x0008) > 0);
	mfd->shiunion		= ((flgs & 0x0010) > 0);

	mfd->nmfunc = IDC_RSH;
	setFlg(mfd);
}

ScriptFunctionDef func_defs[]={
	{ (ScriptFunctionPtr)ScriptConfig, "Config", "0iiiiii" },
		//										   123456
	{ NULL },
};

CScriptObject script_obj={
	NULL, func_defs
};

struct FilterDefinition filterDef_tutorial = {

	NULL, NULL, NULL,		// next, prev, module
	"filter->de- (v0.09)",	// name
	"Just shift chroma or R,G,B channels",
							// desc
	"Eugene Khoroshavin", 		// maker
	NULL,					// private_data
	sizeof(MyFilterData),	// inst_data_size

	InitProc,				// initProc
	NULL,					// deinitProc
	RunProc,				// runProc
	NULL,					// paramProc
	ConfigProc, 			// configProc
	StringProc, 			// stringProc
	StartProc,				// startProc
	EndProc,				// endProc

	&script_obj,			// script_obj
	FssProc,				// fssProc

};

int StartProc(FilterActivation *fa, const FilterFunctions *ff)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	return 0;
}

int EndProc(FilterActivation *fa, const FilterFunctions *ff)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	return 0;
}

///////////////////////////////////////////////////////////////////////////

int RunProc(const FilterActivation *fa, const FilterFunctions *ff)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	const PixDim	width = fa->src.w;
	mfd->width = width;
	const PixDim	height = fa->src.h;
	mfd->height = height;
	Pixel32 *src, *dst;
	int x, y;
	long r, g, b;
	mfd->salign = (fa->src.pitch - width * 4);
	mfd->dalign = (fa->dst.pitch - width * 4);

	src = fa->src.data;
	dst = fa->dst.data;
	for (y = 0; y < height; y++)
	{
		for (x = 0; x < width; x++)
		{
			b = ((UCHAR *)src)[0];
			g = ((UCHAR *)src)[1];
			r = ((UCHAR *)src)[2];

			if (mfd->shiFlg) shiftch(mfd, &r, &g, &b, src, x, y);

			dst[0] = ((int) r << 16) | ((int) g << 8) | ((int) b);
			src++;
			dst++;
		}
		src	= (Pixel32 *)((char *)src + mfd->salign);
		dst	= (Pixel32 *)((char *)dst + mfd->dalign);
	}
	return 0;
}

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_tutorial;

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
	if (!(fd_tutorial = ff->addFilter(fm, &filterDef_tutorial, sizeof(FilterDefinition))))
		return 1;

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

	return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
	ff->removeFilter(fd_tutorial);
}

int InitProc(FilterActivation *fa, const FilterFunctions *ff) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	mfd->rshpt			= 0;
	mfd->rshxx			= 0;
	mfd->rshyy			= 0;
	mfd->gshpt			= 0;
	mfd->gshxx			= 0;
	mfd->gshyy			= 0;
	mfd->bshpt			= 0;
	mfd->bshxx			= 0;
	mfd->bshyy			= 0;

	mfd->keeppt			= TRUE;
	mfd->rshlock		= FALSE;
	mfd->gshlock		= FALSE;
	mfd->bshlock		= FALSE;
	mfd->shiunion		= TRUE;
	mfd->nmfunc			= IDC_RSH;
	return 0;
}

BOOL CALLBACK ConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
	MyFilterData *mfd = (MyFilterData *)GetWindowLong(hdlg, DWL_USER);
	int mPr=0;
	bool redo = false;
	bool undo = false;

	switch(msg) {
		case WM_INITDIALOG:
			SetWindowLong(hdlg, DWL_USER, lParam);
			mfd = (MyFilterData *)lParam;
			HWND hWnd;

			setFlg(mfd);
			hWnd = GetDlgItem(hdlg, IDC_EF1);
			SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-100, 100));
			SendMessage(hWnd, TBM_SETTICFREQ, 10, 0);
			hWnd = GetDlgItem(hdlg, IDC_EF2);
			SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-100, 100));
			SendMessage(hWnd, TBM_SETTICFREQ, 10, 0);
			hWnd = GetDlgItem(hdlg, IDC_EF3);
			SendMessage(hWnd, TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-100, 100));
			SendMessage(hWnd, TBM_SETTICFREQ, 10, 0);

			SendMessage(GetDlgItem(hdlg, IDC_KEEPPT), BM_SETCHECK, mfd->keeppt, 0);
			SendMessage(GetDlgItem(hdlg, IDC_SHIUNION), BM_SETCHECK, mfd->shiunion, 0);
			SendMessage(GetDlgItem(hdlg, IDC_RSHLOCK), BM_SETCHECK, mfd->rshlock, 0);
			SendMessage(GetDlgItem(hdlg, IDC_GSHLOCK), BM_SETCHECK, mfd->gshlock, 0);
			SendMessage(GetDlgItem(hdlg, IDC_BSHLOCK), BM_SETCHECK, mfd->bshlock, 0);

			CheckRadioButton(hdlg,IDC_RSH, IDC_BSH, mfd->nmfunc);
			SendMessage(GetDlgItem(hdlg, mfd->nmfunc), BM_CLICK, TRUE, 0);


			mfd->ifp->InitButton(GetDlgItem(hdlg, IDPREVIEW));
			mfd->ifp->Toggle(hdlg);
			return TRUE;

		case WM_COMMAND:
			switch(LOWORD(wParam)) {
			case IDPREVIEW:
				mfd->ifp->Toggle(hdlg);
				break;
			case IDOK:
				EndDialog(hdlg, 0);
				return TRUE;
			case IDHELP:
				{
				char prog[256];
				char path[256];
				LPTSTR ptr;
				GetModuleFileName(NULL, prog, 255);
				GetFullPathName(prog, 255, path, &ptr);
				*ptr = 0;
				strcat(path, "plugins\\DE.txt");
				OutputDebugString(path);
				OutputDebugString("\n");
				strcpy(prog, "Notepad ");
				strcat(prog, path);
				WinExec(prog, SW_SHOW);
				return TRUE;
				}
			case IDCANCEL:
				EndDialog(hdlg, 1);
				return TRUE;
			case IDC_ZEF1:
				switch (mfd->nmfunc) {
				case IDC_RSH:
					if (mfd->rshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshpt = 0;
							mfd->gshpt = 0;
							mfd->bshpt = 0;
					}
					else mfd->rshpt = 0;
					break;
				case IDC_GSH:
					if (mfd->gshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshpt = 0;
							mfd->gshpt = 0;
							mfd->bshpt = 0;
					}
					else mfd->gshpt = 0;
					break;
				case IDC_BSH:
					if (mfd->bshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshpt = 0;
							mfd->gshpt = 0;
							mfd->bshpt = 0;
					}
					else mfd->bshpt = 0;
					break;
				}
				if (!undo) {
					SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, 0);
					SetDlgItemInt(hdlg, IDC_EF1VAL, 0, TRUE);
					setFlg(mfd);
					mfd->ifp->RedoFrame();
				}
				undo = false;
				return TRUE;
			case IDC_ZEF2:
				switch (mfd->nmfunc) {
				case IDC_RSH:
					if (mfd->rshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshxx = 0;
							mfd->gshxx = 0;
							mfd->bshxx = 0;
					}
					else mfd->rshxx = 0;
					break;
				case IDC_GSH:
					if (mfd->gshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshxx = 0;
							mfd->gshxx = 0;
							mfd->bshxx = 0;
					}
					else mfd->gshxx = 0;
					break;
				case IDC_BSH:
					if (mfd->bshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshxx = 0;
							mfd->gshxx = 0;
							mfd->bshxx = 0;
					}
					else mfd->bshxx = 0;
					break;
				}
				if ( !undo) {
					SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, 0);
					SetDlgItemInt(hdlg, IDC_EF2VAL, 0, TRUE);
					setFlg(mfd);
					mfd->ifp->RedoFrame();
				}
				undo = false;
				return TRUE;
			case IDC_ZEF3:
				switch (mfd->nmfunc) {
				case IDC_RSH:
					if (mfd->rshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshyy = 0;
							mfd->gshyy = 0;
							mfd->bshyy = 0;
					}
					else mfd->rshyy = 0;
					break;
				case IDC_GSH:
					if (mfd->gshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshyy = 0;
							mfd->gshyy = 0;
							mfd->bshyy = 0;
					}
					else mfd->gshyy = 0;
					break;
				case IDC_BSH:
					if (mfd->bshlock) undo = true;
					else if (mfd->shiunion) {
							mfd->rshyy = 0;
							mfd->gshyy = 0;
							mfd->bshyy = 0;
					}
					else mfd->bshyy = 0;
					break;
				}
				if (!undo) {
				SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, 0);
				SetDlgItemInt(hdlg, IDC_EF3VAL, 0, TRUE);
				setFlg(mfd);
				mfd->ifp->RedoFrame();
				}
				undo = false;
				return TRUE;
			case IDC_ZEROALL:
				SendMessage(GetDlgItem(hdlg, IDC_RSHLOCK), BM_SETCHECK, FALSE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_GSHLOCK), BM_SETCHECK, FALSE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_BSHLOCK), BM_SETCHECK, FALSE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_SHIUNION), BM_SETCHECK, TRUE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_KEEPPT), BM_SETCHECK, TRUE, 0);
				mfd->rshlock = FALSE;
				mfd->gshlock = FALSE;
				mfd->bshlock = FALSE;
				mfd->shiunion = TRUE;
				mfd->keeppt = TRUE;

				SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, 0);
				SetDlgItemInt(hdlg, IDC_EF1VAL, 0, TRUE);
				mfd->rshpt = 0;
				mfd->gshpt = 0;
				mfd->bshpt = 0;
				SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, 0);
				SetDlgItemInt(hdlg, IDC_EF2VAL, 0, TRUE);
				mfd->rshxx = 0;
				mfd->gshxx = 0;
				mfd->bshxx = 0;
				SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, 0);
				SetDlgItemInt(hdlg, IDC_EF3VAL, 0, TRUE);
				mfd->rshyy = 0;
				mfd->gshyy = 0;
				mfd->bshyy = 0;
				setFlg(mfd);
				SetDlgItemText(hdlg, IDC_FUNAM, f1ntxt);
				mfd->ifp->RedoFrame();
				return TRUE;
			case IDC_RSH:
				ShowWindow(GetDlgItem(hdlg, IDC_KEEPPT), SW_SHOW);
				SetDlgItemText(hdlg, IDC_EF1TEXT, f1ef1txt);
				SetDlgItemText(hdlg, IDC_EF2TEXT, f1ef2txt);
				SetDlgItemText(hdlg, IDC_EF3TEXT, f1ef3txt);
				if (mfd->shiunion) SetDlgItemText(hdlg, IDC_FUNAM, f1ntxt);
				else SetDlgItemText(hdlg, IDC_FUNAM, f11txt);

				SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, -mfd->rshpt);
				SetDlgItemInt(hdlg, IDC_EF1VAL, mfd->rshpt, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, -mfd->rshxx);
				SetDlgItemInt(hdlg, IDC_EF2VAL, mfd->rshxx, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, -mfd->rshyy);
				SetDlgItemInt(hdlg, IDC_EF3VAL, mfd->rshyy, TRUE);
				mfd->nmfunc = IDC_RSH;
				return TRUE;
			case IDC_GSH:
				ShowWindow(GetDlgItem(hdlg, IDC_KEEPPT), SW_SHOW);
				SetDlgItemText(hdlg, IDC_EF1TEXT, f1ef1txt);
				SetDlgItemText(hdlg, IDC_EF2TEXT, f1ef2txt);
				SetDlgItemText(hdlg, IDC_EF3TEXT, f1ef3txt);
				if (mfd->shiunion) SetDlgItemText(hdlg, IDC_FUNAM, f1ntxt);
				else SetDlgItemText(hdlg, IDC_FUNAM, f12txt);

				SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, -mfd->gshpt);
				SetDlgItemInt(hdlg, IDC_EF1VAL, mfd->gshpt, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, -mfd->gshxx);
				SetDlgItemInt(hdlg, IDC_EF2VAL, mfd->gshxx, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, -mfd->gshyy);
				SetDlgItemInt(hdlg, IDC_EF3VAL, mfd->gshyy, TRUE);
				mfd->nmfunc = IDC_GSH;
				return TRUE;
			case IDC_BSH:
				ShowWindow(GetDlgItem(hdlg, IDC_KEEPPT), SW_SHOW);
				SetDlgItemText(hdlg, IDC_EF1TEXT, f1ef1txt);
				SetDlgItemText(hdlg, IDC_EF2TEXT, f1ef2txt);
				SetDlgItemText(hdlg, IDC_EF3TEXT, f1ef3txt);
				if (mfd->shiunion) SetDlgItemText(hdlg, IDC_FUNAM, f1ntxt);
				else SetDlgItemText(hdlg, IDC_FUNAM, f13txt);
				
				SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, -mfd->bshpt);
				SetDlgItemInt(hdlg, IDC_EF1VAL, mfd->bshpt, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, -mfd->bshxx);
				SetDlgItemInt(hdlg, IDC_EF2VAL, mfd->bshxx, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, -mfd->bshyy);
				SetDlgItemInt(hdlg, IDC_EF3VAL, mfd->bshyy, TRUE);
				mfd->nmfunc = IDC_BSH;
				return TRUE;
			case IDC_KEEPPT: {
				mfd->keeppt = IsDlgButtonChecked(hdlg, IDC_KEEPPT);
				setFlg(mfd);
				mfd->ifp->RedoFrame();}
				return TRUE;
			case IDC_RSHLOCK:
			case IDC_GSHLOCK:
			case IDC_BSHLOCK:
				if (IsDlgButtonChecked(hdlg, LOWORD(wParam))) {
				SendMessage(GetDlgItem(hdlg, IDC_SHIUNION), BM_SETCHECK, FALSE, 0);
				mfd->shiunion = FALSE;
				}
				if (mfd->shiunion) SetDlgItemText(hdlg, IDC_FUNAM, f1ntxt);
				else {
					switch (mfd->nmfunc) {
						case IDC_RSH:
							SetDlgItemText(hdlg, IDC_FUNAM, f11txt);
							break;
						case IDC_GSH:
							SetDlgItemText(hdlg, IDC_FUNAM, f12txt);
							break;
						case IDC_BSH:
							SetDlgItemText(hdlg, IDC_FUNAM, f13txt);
							break;
					}
				}
				mfd->rshlock = IsDlgButtonChecked(hdlg, IDC_RSHLOCK);
				mfd->gshlock = IsDlgButtonChecked(hdlg, IDC_GSHLOCK);
				mfd->bshlock = IsDlgButtonChecked(hdlg, IDC_BSHLOCK);
				return TRUE;
			case IDC_SHIUNION:
				if (mfd->rshlock || mfd->gshlock || mfd->bshlock) {
					SendMessage(GetDlgItem(hdlg, IDC_SHIUNION), BM_SETCHECK, FALSE, 0);
					mfd->shiunion = FALSE;
				}
				else mfd->shiunion = IsDlgButtonChecked(hdlg, IDC_SHIUNION);
				if (mfd->shiunion) SetDlgItemText(hdlg, IDC_FUNAM, f1ntxt);
				else {
					switch (mfd->nmfunc) {
						case IDC_RSH:
							SetDlgItemText(hdlg, IDC_FUNAM, f11txt);
							break;
						case IDC_GSH:
							SetDlgItemText(hdlg, IDC_FUNAM, f12txt);
							break;
						case IDC_BSH:
							SetDlgItemText(hdlg, IDC_FUNAM, f13txt);
							break;
					}
				}
				return TRUE;
			}
			break;
		case WM_VSCROLL:
			{
				int ef1, ef2, ef3, ef1old, ef2old, ef3old, dpos, dw, dh;
				ef1 = -SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_GETPOS, 0, 0);
				ef2 = -SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_GETPOS, 0, 0);
				ef3 = -SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_GETPOS, 0, 0);
				if (mfd->width > 100) dw = 100;
				else dw = mfd->width;
				if (mfd->height > 100) dh = 100;
				else dh = mfd->height;
				switch (mfd->nmfunc) {
				case IDC_RSH:
						ef1old = mfd->rshpt;
						ef2old = mfd->rshxx;
						ef3old = mfd->rshyy;
					if (mfd->rshlock) undo = true;
					else {
						if((mfd->rshpt != ef1) || (mfd->rshxx != ef2) || (mfd->rshyy != ef3)) {
							redo = true;
							if (mfd->shiunion) {
								if (ef1 != ef1old) {
									dpos = ef1 - ef1old;
									dpos = normpos(mfd->rshpt, mfd->gshpt, mfd->bshpt, dpos, dw);
									ef1 = ef1old + dpos;
									mfd->gshpt += dpos;
									mfd->bshpt += dpos;
								}
								if (ef2 != ef2old) {
									dpos = ef2 - ef2old;
									dpos = normpos(mfd->rshxx, mfd->gshxx, mfd->bshxx, dpos, dw);
									ef2 = ef2old + dpos;
									mfd->gshxx += dpos;
									mfd->bshxx += dpos;
								}
								if (ef3 != ef3old) {
									dpos = ef3 - ef3old;
									dpos = normpos(mfd->rshyy, mfd->gshyy, mfd->bshyy, dpos, dh);
									ef3 = ef3old + dpos;
									mfd->gshyy += dpos;
									mfd->bshyy += dpos;
								}
							}
							mfd->rshpt = normv(ef1, dw);
							mfd->rshxx = normv(ef2, dw);
							mfd->rshyy = normv(ef3, dh);
						}
					}
					break;
				case IDC_GSH:
						ef1old = mfd->gshpt;
						ef2old = mfd->gshxx;
						ef3old = mfd->gshyy;
					if (mfd->gshlock) undo = true;
					else {
						if((mfd->gshpt != ef1) || (mfd->gshxx != ef2) || (mfd->gshyy != ef3)) {
							redo = true;
							if (mfd->shiunion) {
								if (ef1 != ef1old) {
									dpos = ef1 - ef1old;
									dpos = normpos(mfd->rshpt, mfd->gshpt, mfd->bshpt, dpos, dw);
									ef1 = ef1old + dpos;
									mfd->rshpt += dpos;
									mfd->bshpt += dpos;
								}
								if (ef2 != ef2old) {
									dpos = ef2 - ef2old;
									dpos = normpos(mfd->rshxx, mfd->gshxx, mfd->bshxx, dpos, dw);
									ef2 = ef2old + dpos;
									mfd->rshxx += dpos;
									mfd->bshxx += dpos;
								}
								if (ef3 != ef3old) {
									dpos = ef3 - ef3old;
									dpos = normpos(mfd->rshyy, mfd->gshyy, mfd->bshyy, dpos, dh);
									ef3 = ef3old + dpos;
									mfd->rshyy += dpos;
									mfd->bshyy += dpos;
								}
							}
							mfd->gshpt = normv(ef1, dw);
							mfd->gshxx = normv(ef2, dw);
							mfd->gshyy = normv(ef3, dh);
						}
					}
					break;
				case IDC_BSH:
						ef1old = mfd->bshpt;
						ef2old = mfd->bshxx;
						ef3old = mfd->bshyy;
					if (mfd->bshlock) undo = true;
					else {
						if((mfd->bshpt != ef1) || (mfd->bshxx != ef2) || (mfd->bshyy != ef3)) {
							redo = true;
							if (mfd->shiunion) {
								if (ef1 != ef1old) {
									dpos = ef1 - ef1old;
									dpos = normpos(mfd->rshpt, mfd->gshpt, mfd->bshpt, dpos, dw);
									ef1 = ef1old + dpos;
									mfd->rshpt += dpos;
									mfd->gshpt += dpos;
								}
								if (ef2 != ef2old) {
									dpos = ef2 - ef2old;
									dpos = normpos(mfd->rshxx, mfd->gshxx, mfd->bshxx, dpos, dw);
									ef2 = ef2old + dpos;
									mfd->rshxx += dpos;
									mfd->gshxx += dpos;
								}
								if (ef3 != ef3old) {
									dpos = ef3 - ef3old;
									dpos = normpos(mfd->rshyy, mfd->gshyy, mfd->bshyy, dpos, dh);
									ef3 = ef3old + dpos;
									mfd->rshyy += dpos;
									mfd->gshyy += dpos;
								}
							}
							mfd->bshpt = normv(ef1, dw);
							mfd->bshxx = normv(ef2, dw);
							mfd->bshyy = normv(ef3, dh);
						}
					}
					break;
				}
				if (undo) {
					SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, -ef1old);
					SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, -ef2old);
					SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, -ef3old);
				}
				if (redo) {
					SendMessage(GetDlgItem(hdlg, IDC_EF1), TBM_SETPOS, (WPARAM)TRUE, -ef1);
					SendMessage(GetDlgItem(hdlg, IDC_EF2), TBM_SETPOS, (WPARAM)TRUE, -ef2);
					SendMessage(GetDlgItem(hdlg, IDC_EF3), TBM_SETPOS, (WPARAM)TRUE, -ef3);
					SetDlgItemInt(hdlg, IDC_EF1VAL, ef1, TRUE);
					SetDlgItemInt(hdlg, IDC_EF2VAL, ef2, TRUE);
					SetDlgItemInt(hdlg, IDC_EF3VAL, ef3, TRUE);
					setFlg(mfd);
					mfd->ifp->RedoFrame();
				}
				undo = false;
				redo = false;
				break;
			}
	}

	return FALSE;
}

int ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	MyFilterData mfd_old = *mfd;
	int ret;

	mfd->ifp = fa->ifp;

	if (DialogBoxParam(fa->filter->module->hInstModule, MAKEINTRESOURCE(IDD_FILTER),
		hwnd, ConfigDlgProc, (LPARAM)mfd))
	{
		*mfd = mfd_old;
		ret = TRUE;
	}
    else
	{
		ret = FALSE;
	}
	return(ret);
}

void StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
}
 

void setFlg ( MyFilterData *mfd ) {
	mfd->shiFlg =(
			(mfd->rshpt != 0) || (mfd->gshpt != 0) || (mfd->bshpt != 0)
		||	(mfd->rshxx != 0) || (mfd->gshxx != 0) || (mfd->bshxx != 0)
		||	(mfd->rshyy != 0) || (mfd->gshyy != 0) || (mfd->bshyy != 0));
		switch (mfd->nmfunc) {
		case IDC_RSH:
		case IDC_GSH:
		case IDC_BSH:
			mfd->rshxy = mfd->rshyy * (mfd->width * 4 + mfd->salign) + mfd->rshxx * 4 + mfd->rshpt * 4;
			mfd->gshxy = mfd->gshyy * (mfd->width * 4 + mfd->salign) + mfd->gshxx * 4 + mfd->gshpt * 4;
			mfd->bshxy = mfd->bshyy * (mfd->width * 4 + mfd->salign) + mfd->bshxx * 4 + mfd->bshpt * 4;
			break;

	}
}


int normpos( int a, int b, int c, int dpos, long width) {
int ret;
if (dpos >0) {
		ret = a;
		if (b > ret) ret = b;
		if (c > ret) ret = c;
		if ((ret + dpos) < (width + 1)) ret = dpos;
		else ret = width - ret;
}
else {
		ret = a;
		if (b < ret) ret = b;
		if (c < ret) ret = c;
		if ((ret + dpos) > -(width + 1)) ret = dpos;
		else ret = -width - ret;
}
return(ret);
}

inline void normc(long *c) {
	if (*c < 0) *c = 0;
	if (*c > 255) *c = 255;
}

inline void normrgb(long *r, long *g, long *b) {
	if (*r < 0) *r = 0;
	if (*r > 255) *r = 255;
	if (*g < 0) *g = 0;
	if (*g > 255) *g = 255;
	if (*b < 0) *b = 0;
	if (*b > 255) *b = 255;
}


inline void shiftch(MyFilterData *mfd, long *r, long *g, long *b, Pixel32 *src,  int x, int y) {
	long udp, ri, gi, bi, uri, ugi, ubi;
	UCHAR *srcl;
	udp = (*r + *g + *b)/3;
	bool nmiddle, pre, post;

	pre = (x < (mfd->rshxx + mfd->rshpt));
	post = (x >= (mfd->width + mfd->rshxx + mfd->rshpt));
	nmiddle = (y < mfd->rshyy)
		|| (y >= (mfd->height + mfd->rshyy))
		|| (x < mfd->rshxx)
		|| (x >= (mfd->width + mfd->rshxx))
		|| (pre && (y <= mfd->rshyy))
		|| (post && (y >= (mfd->height - 1 + mfd->rshyy)));
	if (nmiddle) {
		ri = udp;
		uri = ri;
	}
	else {
		srcl = (UCHAR *)src - mfd->rshxy;
		if (pre) srcl -= mfd->salign;
		if (post) srcl += mfd->salign;
		ri = srcl[2];
		uri = (ri + srcl[1] + srcl[0])/3;
	}

	pre = (x < (mfd->gshxx + mfd->gshpt));
	post = (x >= (mfd->width + mfd->gshxx + mfd->gshpt));
	nmiddle = (y < mfd->gshyy)
		|| (y >= (mfd->height + mfd->gshyy))
		|| (x < mfd->gshxx)
		|| (x >= (mfd->width + mfd->gshxx))
		|| (pre && (y <= mfd->gshyy))
		|| (post && (y >= (mfd->height - 1 + mfd->gshyy)));
	if (nmiddle) {
		gi = udp;
		ugi = gi;
	}
	else {
		srcl = (UCHAR *)src - mfd->gshxy;
		if (pre) srcl -=mfd->salign;
		if (post) srcl +=mfd->salign;
		gi = srcl[1];
		ugi = (srcl[2] + gi + srcl[0])/3;
	}

	pre = (x < (mfd->bshxx + mfd->bshpt));
	post = (x >= (mfd->width + mfd->bshxx + mfd->bshpt));
	nmiddle = (y < mfd->bshyy)
		|| (y >= (mfd->height + mfd->bshyy))
		|| (x < mfd->bshxx)
		|| (x >= (mfd->width + mfd->bshxx))
		|| (pre && (y <= mfd->bshyy))
		|| (post && (y >= (mfd->height - 1 + mfd->bshyy)));
	if (nmiddle) {
		bi = udp;
		ubi = bi;
	}
	else {
		srcl = (UCHAR *)src - mfd->bshxy;
		if (pre) srcl -= mfd->salign;
		if (post) srcl += mfd->salign;
		bi = srcl[0];
		ubi = (srcl[2] + srcl[1] + bi)/3;
	}

	if (mfd->keeppt) {
		*r = ri - uri + udp;
		*g = gi - ugi + udp;
		*b = bi - ubi + udp;
		}
	else {
		*r = ri;
		*g = gi;
		*b = bi;
	}
	normrgb(r, g, b);
}

inline int normv(int v, int norm) {
if ( v > norm) return norm;
else return v;
if ( v < -norm) return -norm;
else return v;
}